@@ -18,6 +18,7 @@ import csv as csv_std
 import tabulate
 from .utils import (
     _compile_code,
+    chunks,
     file_progress,
     find_spatialite,
     sqlite3,
@@ -28,6 +29,7 @@ from .utils import (
     TypeTracker,
 )
 
+
 CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
 
 VALID_COLUMN_TYPES = ("INTEGER", "TEXT", "FLOAT", "BLOB")
@@ -75,7 +77,9 @@ def output_options(fn):
             click.option("--csv", is_flag=True, help="Output CSV"),
             click.option("--tsv", is_flag=True, help="Output TSV"),
             click.option("--no-headers", is_flag=True, help="Omit CSV headers"),
-            click.option("-t", "--table", is_flag=True, help="Output as a table"),
+            click.option(
+                "-t", "--table", is_flag=True, help="Output as a formatted table"
+            ),
             click.option(
                 "--fmt",
                 help="Table format - one of {}".format(
@@ -161,7 +165,13 @@ def tables(
     load_extension,
     views=False,
 ):
-    """List the tables in the database"""
+    """List the tables in the database
+
+    Example:
+
+    \b
+        sqlite-utils tables trees.db
+    """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
     headers = ["view" if views else "table"]
@@ -242,7 +252,13 @@ def views(
     schema,
     load_extension,
 ):
-    """List the views in the database"""
+    """List the views in the database
+
+    Example:
+
+    \b
+        sqlite-utils views trees.db
+    """
     tables.callback(
         path=path,
         fts4=False,
@@ -273,7 +289,13 @@ def views(
 @click.option("--no-vacuum", help="Don't run VACUUM", default=False, is_flag=True)
 @load_extension_option
 def optimize(path, tables, no_vacuum, load_extension):
-    """Optimize all full-text search tables and then run VACUUM - should shrink the database file"""
+    """Optimize all full-text search tables and then run VACUUM - should shrink the database file
+
+    Example:
+
+    \b
+        sqlite-utils optimize chickens.db
+    """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
     if not tables:
@@ -294,7 +316,13 @@ def optimize(path, tables, no_vacuum, load_extension):
 @click.argument("tables", nargs=-1)
 @load_extension_option
 def rebuild_fts(path, tables, load_extension):
-    """Rebuild all or specific full-text search tables"""
+    """Rebuild all or specific full-text search tables
+
+    Example:
+
+    \b
+        sqlite-utils rebuild-fts chickens.db chickens
+    """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
     if not tables:
@@ -312,7 +340,13 @@ def rebuild_fts(path, tables, load_extension):
 )
 @click.argument("names", nargs=-1)
 def analyze(path, names):
-    """Run ANALYZE against the whole database, or against specific named indexes and tables"""
+    """Run ANALYZE against the whole database, or against specific named indexes and tables
+
+    Example:
+
+    \b
+        sqlite-utils analyze chickens.db
+    """
     db = sqlite_utils.Database(path)
     try:
         if names:
@@ -331,7 +365,13 @@ def analyze(path, names):
     required=True,
 )
 def vacuum(path):
-    """Run VACUUM against the database"""
+    """Run VACUUM against the database
+
+    Example:
+
+    \b
+        sqlite-utils vacuum chickens.db
+    """
     sqlite_utils.Database(path).vacuum()
 
 
@@ -343,7 +383,13 @@ def vacuum(path):
 )
 @load_extension_option
 def dump(path, load_extension):
-    """Output a SQL dump of the schema and full contents of the database"""
+    """Output a SQL dump of the schema and full contents of the database
+
+    Example:
+
+    \b
+        sqlite-utils dump chickens.db
+    """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
     for line in db.conn.iterdump():
@@ -384,7 +430,13 @@ def dump(path, load_extension):
 def add_column(
     path, table, col_name, col_type, fk, fk_col, not_null_default, load_extension
 ):
-    "Add a column to the specified table"
+    """Add a column to the specified table
+
+    Example:
+
+    \b
+        sqlite-utils add-column chickens.db chickens weight float
+    """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
     db[table].add_column(
@@ -412,9 +464,11 @@ def add_foreign_key(
     path, table, column, other_table, other_column, ignore, load_extension
 ):
     """
-    Add a new foreign key constraint to an existing table. Example usage:
+    Add a new foreign key constraint to an existing table
 
-        $ sqlite-utils add-foreign-key my.db books author_id authors id
+    Example:
+
+        sqlite-utils add-foreign-key my.db books author_id authors id
 
     WARNING: Could corrupt your database! Back up your database file first.
     """
@@ -436,12 +490,14 @@ def add_foreign_key(
 @load_extension_option
 def add_foreign_keys(path, foreign_key, load_extension):
     """
-    Add multiple new foreign key constraints to a database. Example usage:
+    Add multiple new foreign key constraints to a database
+
+    Example:
 
     \b
-    sqlite-utils add-foreign-keys my.db \\
-        books author_id authors id \\
-        authors country_id countries id
+        sqlite-utils add-foreign-keys my.db \\
+            books author_id authors id \\
+            authors country_id countries id
     """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
@@ -467,7 +523,12 @@ def add_foreign_keys(path, foreign_key, load_extension):
 @load_extension_option
 def index_foreign_keys(path, load_extension):
     """
-    Ensure every foreign key column has an index on it.
+    Ensure every foreign key column has an index on it
+
+    Example:
+
+    \b
+        sqlite-utils index-foreign-keys chickens.db
     """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
@@ -500,9 +561,17 @@ def create_index(
     path, table, column, name, unique, if_not_exists, analyze, load_extension
 ):
     """
-    Add an index to the specified table covering the specified columns.
-    Use "sqlite-utils create-index mydb -- -column" to specify descending
-    order for a column.
+    Add an index to the specified table for the specified columns
+
+    Example:
+
+    \b
+        sqlite-utils create-index chickens.db chickens name
+
+    To create an index in descending order:
+
+    \b
+        sqlite-utils create-index chickens.db chickens -- -name
     """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
@@ -542,7 +611,13 @@ def create_index(
 def enable_fts(
     path, table, column, fts4, fts5, tokenize, create_triggers, load_extension
 ):
-    "Enable full-text search for specific table and columns"
+    """Enable full-text search for specific table and columns"
+
+    Example:
+
+    \b
+        sqlite-utils enable-fts chickens.db chickens name
+    """
     fts_version = "FTS5"
     if fts4 and fts5:
         click.echo("Can only use one of --fts4 or --fts5", err=True)
@@ -570,7 +645,13 @@ def enable_fts(
 @click.argument("column", nargs=-1, required=True)
 @load_extension_option
 def populate_fts(path, table, column, load_extension):
-    "Re-populate full-text search for specific table and columns"
+    """Re-populate full-text search for specific table and columns
+
+    Example:
+
+    \b
+        sqlite-utils populate-fts chickens.db chickens name
+    """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
     db[table].populate_fts(column)
@@ -585,7 +666,13 @@ def populate_fts(path, table, column, load_extension):
 @click.argument("table")
 @load_extension_option
 def disable_fts(path, table, load_extension):
-    "Disable full-text search for specific table"
+    """Disable full-text search for specific table
+
+    Example:
+
+    \b
+        sqlite-utils disable-fts chickens.db chickens
+    """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
     db[table].disable_fts()
@@ -600,7 +687,13 @@ def disable_fts(path, table, load_extension):
 )
 @load_extension_option
 def enable_wal(path, load_extension):
-    "Enable WAL for database files"
+    """Enable WAL for database files
+
+    Example:
+
+    \b
+        sqlite-utils enable-wal chickens.db
+    """
     for path_ in path:
         db = sqlite_utils.Database(path_)
         _load_extensions(db, load_extension)
@@ -616,7 +709,13 @@ def enable_wal(path, load_extension):
 )
 @load_extension_option
 def disable_wal(path, load_extension):
-    "Disable WAL for database files"
+    """Disable WAL for database files
+
+    Example:
+
+    \b
+        sqlite-utils disable-wal chickens.db
+    """
     for path_ in path:
         db = sqlite_utils.Database(path_)
         _load_extensions(db, load_extension)
@@ -632,7 +731,13 @@ def disable_wal(path, load_extension):
 @click.argument("tables", nargs=-1)
 @load_extension_option
 def enable_counts(path, tables, load_extension):
-    "Configure triggers to update a _counts table with row counts"
+    """Configure triggers to update a _counts table with row counts
+
+    Example:
+
+    \b
+        sqlite-utils enable-counts chickens.db
+    """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
     if not tables:
@@ -654,7 +759,13 @@ def enable_counts(path, tables, load_extension):
 )
 @load_extension_option
 def reset_counts(path, load_extension):
-    "Reset calculated counts in the _counts table"
+    """Reset calculated counts in the _counts table
+
+    Example:
+
+    \b
+        sqlite-utils reset-counts chickens.db
+    """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
     db.reset_counts()
@@ -704,59 +815,65 @@ def import_options(fn):
     return fn
 
 
-def insert_upsert_options(fn):
-    for decorator in reversed(
-        (
-            click.argument(
-                "path",
-                type=click.Path(file_okay=True, dir_okay=False, allow_dash=False),
-                required=True,
-            ),
-            click.argument("table"),
-            click.argument("file", type=click.File("rb"), required=True),
-            click.option(
-                "--pk", help="Columns to use as the primary key, e.g. id", multiple=True
-            ),
-        )
-        + _import_options
-        + (
-            click.option(
-                "--batch-size", type=int, default=100, help="Commit every X records"
-            ),
-            click.option(
-                "--alter",
-                is_flag=True,
-                help="Alter existing table to add any missing columns",
-            ),
-            click.option(
-                "--not-null",
-                multiple=True,
-                help="Columns that should be created as NOT NULL",
-            ),
-            click.option(
-                "--default",
-                multiple=True,
-                type=(str, str),
-                help="Default value that should be set for a column",
-            ),
-            click.option(
-                "-d",
-                "--detect-types",
-                is_flag=True,
-                envvar="SQLITE_UTILS_DETECT_TYPES",
-                help="Detect types for columns in CSV/TSV data",
-            ),
-            click.option(
-                "--analyze",
-                is_flag=True,
-                help="Run ANALYZE at the end of this operation",
-            ),
-            load_extension_option,
-            click.option("--silent", is_flag=True, help="Do not show progress bar"),
-        )
-    ):
-        fn = decorator(fn)
-    return fn
+def insert_upsert_options(*, require_pk=False):
+    def inner(fn):
+        for decorator in reversed(
+            (
+                click.argument(
+                    "path",
+                    type=click.Path(file_okay=True, dir_okay=False, allow_dash=False),
+                    required=True,
+                ),
+                click.argument("table"),
+                click.argument("file", type=click.File("rb"), required=True),
+                click.option(
+                    "--pk",
+                    help="Columns to use as the primary key, e.g. id",
+                    multiple=True,
+                    required=require_pk,
+                ),
+            )
+            + _import_options
+            + (
+                click.option(
+                    "--batch-size", type=int, default=100, help="Commit every X records"
+                ),
+                click.option(
+                    "--alter",
+                    is_flag=True,
+                    help="Alter existing table to add any missing columns",
+                ),
+                click.option(
+                    "--not-null",
+                    multiple=True,
+                    help="Columns that should be created as NOT NULL",
+                ),
+                click.option(
+                    "--default",
+                    multiple=True,
+                    type=(str, str),
+                    help="Default value that should be set for a column",
+                ),
+                click.option(
+                    "-d",
+                    "--detect-types",
+                    is_flag=True,
+                    envvar="SQLITE_UTILS_DETECT_TYPES",
+                    help="Detect types for columns in CSV/TSV data",
+                ),
+                click.option(
+                    "--analyze",
+                    is_flag=True,
+                    help="Run ANALYZE at the end of this operation",
+                ),
+                load_extension_option,
+                click.option("--silent", is_flag=True, help="Do not show progress bar"),
+            )
+        ):
+            fn = decorator(fn)
+        return fn
+
+    return inner
 
 
 def insert_upsert_implementation(
@@ -893,13 +1010,21 @@ def insert_upsert_implementation(
     if upsert:
         extra_kwargs["upsert"] = upsert
 
+    # docs should all be dictionaries
+    docs = (verify_is_dict(doc) for doc in docs)
+
     # Apply {"$base64": true, ...} decoding, if needed
     docs = (decode_base64_values(doc) for doc in docs)
 
     # For bulk_sql= we use cursor.executemany() instead
     if bulk_sql:
-        with db.conn:
-            db.conn.cursor().executemany(bulk_sql, docs)
+        if batch_size:
+            doc_chunks = chunks(docs, batch_size)
+        else:
+            doc_chunks = [docs]
+        for doc_chunk in doc_chunks:
+            with db.conn:
+                db.conn.cursor().executemany(bulk_sql, doc_chunk)
         return
 
     try:
@@ -951,7 +1076,7 @@ def _find_variables(tb, vars):
 
 
 @cli.command()
-@insert_upsert_options
+@insert_upsert_options()
 @click.option(
     "--ignore", is_flag=True, default=False, help="Ignore records if pk already exists"
 )
@@ -1001,7 +1126,11 @@ def insert(
     Insert records from FILE into a table, creating the table if it
     does not already exist.
 
-    By default the input is expected to be a JSON array of objects. Or:
+    Example:
+
+        echo '{"name": "Lila"}' | sqlite-utils insert data.db chickens -
+
+    By default the input is expected to be a JSON object or array of objects.
 
     \b
     - Use --nl for newline-delimited JSON objects
@@ -1055,7 +1184,7 @@ def insert(
 
 
 @cli.command()
-@insert_upsert_options
+@insert_upsert_options(require_pk=True)
 def upsert(
     path,
     table,
@@ -1087,6 +1216,14 @@ def upsert(
     Upsert records based on their primary key. Works like 'insert' but if
     an incoming record has a primary key that matches an existing record
     the existing record will be updated.
+
+    Example:
+
+    \b
+        echo '[
+            {"id": 1, "name": "Lila"},
+            {"id": 2, "name": "Suna"}
+        ]' | sqlite-utils upsert data.db chickens - --pk id
     """
     try:
         insert_upsert_implementation(
@@ -1129,12 +1266,14 @@ def upsert(
 )
 @click.argument("sql")
 @click.argument("file", type=click.File("rb"), required=True)
+@click.option("--batch-size", type=int, default=100, help="Commit every X records")
 @import_options
 @load_extension_option
 def bulk(
     path,
-    file,
     sql,
+    file,
+    batch_size,
     flatten,
     nl,
     csv,
@@ -1152,6 +1291,16 @@ def bulk(
 ):
     """
     Execute parameterized SQL against the provided list of documents.
+
+    Example:
+
+    \b
+        echo '[
+            {"id": 1, "name": "Lila2"},
+            {"id": 2, "name": "Suna2"}
+        ]' | sqlite-utils bulk data.db '
+            update chickens set name = :name where id = :id
+        ' -
     """
     try:
         insert_upsert_implementation(
@@ -1172,7 +1321,7 @@ def bulk(
             sniff=sniff,
             no_headers=no_headers,
             encoding=encoding,
-            batch_size=1,
+            batch_size=batch_size,
             alter=False,
             upsert=False,
             not_null=set(),
@@ -1196,7 +1345,13 @@ def bulk(
     "--enable-wal", is_flag=True, help="Enable WAL mode on the created database"
 )
 def create_database(path, enable_wal):
-    "Create a new empty database file."
+    """Create a new empty database file
+
+    Example:
+
+    \b
+        sqlite-utils create-database trees.db
+    """
     db = sqlite_utils.Database(path)
     if enable_wal:
         db.enable_wal()
@@ -1248,11 +1403,11 @@ def create_table(
     name, type pairs, for example:
 
     \b
-    sqlite-utils create-table my.db people \\
-        id integer \\
-        name text \\
-        height float \\
-        photo blob --pk id
+        sqlite-utils create-table my.db people \\
+            id integer \\
+            name text \\
+            height float \\
+            photo blob --pk id
     """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
@@ -1297,7 +1452,13 @@ def create_table(
 @click.option("--ignore", is_flag=True)
 @load_extension_option
 def drop_table(path, table, ignore, load_extension):
-    "Drop the specified table"
+    """Drop the specified table
+
+    Example:
+
+    \b
+        sqlite-utils drop-table chickens.db chickens
+    """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
     try:
@@ -1326,7 +1487,14 @@ def drop_table(path, table, ignore, load_extension):
 )
 @load_extension_option
 def create_view(path, view, select, ignore, replace, load_extension):
-    "Create a view for the provided SELECT query"
+    """Create a view for the provided SELECT query
+
+    Example:
+
+    \b
+        sqlite-utils create-view chickens.db heavy_chickens \\
+          'select * from chickens where weight > 3'
+    """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
     # Does view already exist?
@@ -1354,7 +1522,13 @@ def create_view(path, view, select, ignore, replace, load_extension):
 @click.option("--ignore", is_flag=True)
 @load_extension_option
 def drop_view(path, view, ignore, load_extension):
-    "Drop the specified view"
+    """Drop the specified view
+
+    Example:
+
+    \b
+        sqlite-utils drop-view chickens.db heavy_chickens
+    """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
     try:
@@ -1402,7 +1576,15 @@ def query(
     param,
     load_extension,
 ):
-    "Execute SQL query and return the results as JSON"
+    """Execute SQL query and return the results as JSON
+
+    Example:
+
+    \b
+        sqlite-utils data.db \\
+            "select * from chickens where age > :age" \\
+            -p age 1
+    """
     db = sqlite_utils.Database(path)
     for alias, attach_path in attach:
         db.attach(alias, attach_path)
@@ -1665,7 +1847,12 @@ def search(
     json_cols,
     load_extension,
 ):
-    "Execute a full-text search against this table"
+    """Execute a full-text search against this table
+
+    Example:
+
+        sqlite-utils search data.db chickens lila
+    """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
     # Check table exists
@@ -1763,7 +1950,13 @@ def rows(
     json_cols,
     load_extension,
 ):
-    "Output all rows in the specified table"
+    """Output all rows in the specified table
+
+    Example:
+
+    \b
+        sqlite-utils rows trees.db Trees
+    """
     columns = "*"
     if column:
         columns = ", ".join("[{}]".format(c) for c in column)
@@ -1815,7 +2008,13 @@ def triggers(
     json_cols,
     load_extension,
 ):
-    "Show triggers configured in this database"
+    """Show triggers configured in this database
+
+    Example:
+
+    \b
+        sqlite-utils triggers trees.db
+    """
     sql = "select name, tbl_name as [table], sql from sqlite_master where type = 'trigger'"
     if tables:
         quote = sqlite_utils.Database(memory=True).quote
@@ -1864,7 +2063,13 @@ def indexes(
     json_cols,
     load_extension,
 ):
-    "Show indexes for this database"
+    """Show indexes for the whole database or specific tables
+
+    Example:
+
+    \b
+        sqlite-utils indexes trees.db Trees
+    """
     sql = """
     select
       sqlite_master.name as "table",
@@ -1912,7 +2117,13 @@ def schema(
     tables,
     load_extension,
 ):
-    "Show full schema for this database or for specified tables"
+    """Show full schema for this database or for specified tables
+
+    Example:
+
+    \b
+        sqlite-utils schema trees.db
+    """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
     if tables:
@@ -1964,7 +2175,7 @@ def schema(
     "--drop-foreign-key",
     type=str,
     multiple=True,
-    help="Drop this foreign key constraint",
+    help="Drop foreign key constraint for this column",
 )
 @click.option("--sql", is_flag=True, help="Output SQL without executing it")
 @load_extension_option
@@ -1985,7 +2196,15 @@ def transform(
     sql,
     load_extension,
 ):
-    "Transform a table beyond the capabilities of ALTER TABLE"
+    """Transform a table beyond the capabilities of ALTER TABLE
+
+    Example:
+
+    \b
+        sqlite-utils transform mydb.db mytable \\
+            --drop column1 \\
+            --rename column2 column_renamed
+    """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
     types = {}
@@ -2060,7 +2279,13 @@ def extract(
     rename,
     load_extension,
 ):
-    "Extract one or more columns into a separate table"
+    """Extract one or more columns into a separate table
+
+    Example:
+
+    \b
+        sqlite-utils extract trees.db Street_Trees species
+    """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
     kwargs = dict(
@@ -2122,17 +2347,17 @@ def insert_files(
     """
     Insert one or more files using BLOB columns in the specified table
 
-    Example usage:
+    Example:
 
     \b
-    sqlite-utils insert-files pics.db images *.gif \\
-        -c name:name \\
-        -c content:content \\
-        -c content_hash:sha256 \\
-        -c created:ctime_iso \\
-        -c modified:mtime_iso \\
-        -c size:size \\
-        --pk name
+        sqlite-utils insert-files pics.db images *.gif \\
+            -c name:name \\
+            -c content:content \\
+            -c content_hash:sha256 \\
+            -c created:ctime_iso \\
+            -c modified:mtime_iso \\
+            -c size:size \\
+            --pk name
     """
     if not column:
         if text:
@@ -2244,7 +2469,13 @@ def analyze_tables(
     save,
     load_extension,
 ):
-    "Analyze the columns in one or more tables"
+    """Analyze the columns in one or more tables
+
+    Example:
+
+    \b
+        sqlite-utils analyze-tables data.db trees
+    """
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
     _analyze(db, tables, columns, save)
@@ -2308,9 +2539,9 @@ def _generate_convert_help():
     Convert columns using Python code you supply. For example:
 
     \b
-    $ sqlite-utils convert my.db mytable mycolumn \\
-        '"\\n".join(textwrap.wrap(value, 10))' \\
-        --import=textwrap
+        sqlite-utils convert my.db mytable mycolumn \\
+            '"\\n".join(textwrap.wrap(value, 10))' \\
+            --import=textwrap
 
     "value" is a variable with the column value to be converted.
 
@@ -2333,8 +2564,8 @@ def _generate_convert_help():
     You can use these recipes like so:
 
     \b
-    $ sqlite-utils convert my.db mytable mycolumn \\
-        'r.jsonsplit(value, delimiter=":")'
+        sqlite-utils convert my.db mytable mycolumn \\
+            'r.jsonsplit(value, delimiter=":")'
     """
     ).strip()
     return help
@@ -2533,6 +2764,14 @@ def json_binary(value):
         raise TypeError
 
 
+def verify_is_dict(doc):
+    if not isinstance(doc, dict):
+        raise click.ClickException(
+            "Rows must all be dictionaries, got: {}".format(repr(doc)[:1000])
+        )
+    return doc
+
+
 def _load_extensions(db, load_extension):
     if load_extension:
         db.conn.enable_load_extension(True)
