@@ -774,6 +774,17 @@ def test_create_index_find_unique_name(fresh_db):
     assert index_names == {"idx_t_id", "idx_t_id_2", "idx_t_id_3"}
 
 
+def test_create_index_analyze(fresh_db):
+    dogs = fresh_db["dogs"]
+    assert "sqlite_stat1" not in fresh_db.table_names()
+    dogs.insert({"name": "Cleo", "twitter": "cleopaws"})
+    dogs.create_index(["name"], analyze=True)
+    assert "sqlite_stat1" in fresh_db.table_names()
+    assert list(fresh_db["sqlite_stat1"].rows) == [
+        {"tbl": "dogs", "idx": "idx_dogs_name", "stat": "1 1"}
+    ]
+
+
 @pytest.mark.parametrize(
     "data_structure",
     (
@@ -1017,6 +1028,23 @@ def test_insert_all_single_column(fresh_db):
     assert table.pks == ["name"]
 
 
+@pytest.mark.parametrize("method_name", ("insert_all", "upsert_all"))
+def test_insert_all_analyze(fresh_db, method_name):
+    table = fresh_db["table"]
+    table.insert_all([{"id": 1, "name": "Cleo"}], pk="id")
+    assert "sqlite_stat1" not in fresh_db.table_names()
+    table.create_index(["name"], analyze=True)
+    assert list(fresh_db["sqlite_stat1"].rows) == [
+        {"tbl": "table", "idx": "idx_table_name", "stat": "1 1"}
+    ]
+    method = getattr(table, method_name)
+    method([{"id": 2, "name": "Suna"}], pk="id", analyze=True)
+    assert "sqlite_stat1" in fresh_db.table_names()
+    assert list(fresh_db["sqlite_stat1"].rows) == [
+        {"tbl": "table", "idx": "idx_table_name", "stat": "2 1"}
+    ]
+
+
 def test_create_with_a_null_column(fresh_db):
     record = {"name": "Name", "description": None}
     fresh_db["t"].insert(record)
